
import hashlib
from fileio.core.types import FileType, get_filelike

DEFAULT_CHUNK_SIZE = 1024

def checksum_file(
    path: FileType, 
    chunk_size: int = DEFAULT_CHUNK_SIZE
):
    """
    Takes the uploaded file from the request and performs
    a SHA256 checksum to determine the uniqueness of the file.
    """
    path = get_filelike(path)
    sha256_hash = hashlib.sha256()
    with path.open('rb') as f:
        for byte_block in iter(lambda: f.read(chunk_size), b""):
            sha256_hash.update(byte_block)
    checksum = sha256_hash.hexdigest()
    del sha256_hash
    return checksum

async def async_checksum_file(
    path: FileType, 
    chunk_size: int = DEFAULT_CHUNK_SIZE
):
    """
    [Async]
    Takes the uploaded file from the request and performs
    a SHA256 checksum to determine the uniqueness of the file.
    """
    path = get_filelike(path)
    sha256_hash = hashlib.sha256()
    async with path.async_open('rb') as f:
        for byte_block in await iter(lambda: f.read(chunk_size), b""):
            sha256_hash.update(byte_block)
    checksum = sha256_hash.hexdigest()
    del sha256_hash
    return checksum


def checksum_bytes(data: bytes):
    """
    Takes the bytes of a file and performs
    a SHA256 checksum to determine the uniqueness of the file.
    """
    sha256_hash = hashlib.sha256()
    sha256_hash.update(data)
    checksum = sha256_hash.hexdigest()
    del sha256_hash
    return checksum



from __future__ import absolute_import

from typing import Dict
from fileio.io._base import BasePack
from fileio.io._base64 import Base64
from fileio.io._base64gzip import Base64GZip
from fileio.io._json import Json, OrJson
from fileio.io._pickle import Pickle, Dill
from fileio.io._msgpack import MsgPack
from fileio.io._generator import Generate
from fileio.io._text import Text
from fileio.io._yaml import Yaml
from fileio.io._csv import Csv
from fileio.io._tsv import Tsv
from fileio.io._jsonlines import JsonLines

Serializers: Dict[str, BasePack] = {
    'pickle': Pickle,
    'dill': Dill,
    'json': Json,
    'orjson': OrJson,
    'base64': Base64,
    'bgzip': Base64GZip,
    'msgpack': MsgPack,
    'yaml': Yaml,
    'text': Text,
}




from __future__ import annotations

import functools
from typing import AsyncIterable, Callable, List, TYPE_CHECKING

from fileio.core._pathlib import _is_wildcard_pattern, _ignore_error, _Flavour
from fileio.aiopath.wrap import CoroutineMethod

if TYPE_CHECKING:  # keep mypy quiet
  from fileio.core.base import FilePath

class _AsyncSelector:
    """A selector matches a specific glob pattern part against the children
        of a given path.
    """

    def __init__(self, child_parts: List[str], flavour: _Flavour):
        self.child_parts: List[str] = child_parts

        if child_parts:
            self.successor = _make_selector(child_parts, flavour)
            self.dironly = True

        else:
            self.successor = _TerminatingSelector()
            self.dironly = False

    async def select_from(
        self,
        parent_path: 'FilePath'
    ) -> AsyncIterable['FilePath']:
        """Iterate over all child paths of `parent_path` matched by this
        selector.  This can contain parent_path itself."""
        path_cls = type(parent_path)
        is_dir = path_cls.is_dir
        exists = path_cls.exists
        scandir = parent_path._accessor.scandir

        if not await is_dir(parent_path):
            return

        async for item in self._select_from(parent_path, is_dir, exists, scandir):
            yield item


class _TerminatingSelector:
    async def _select_from(
        self,
        parent_path: 'FilePath', 
        is_dir: CoroutineMethod, 
        exists: CoroutineMethod, 
        scandir: CoroutineMethod
    ) -> AsyncIterable['FilePath']:
        yield parent_path


class _PreciseSelector(_AsyncSelector):
    def __init__(self, name: str, child_parts: List[str], flavour: _Flavour):
        self.name: str = name
        super().__init__(child_parts, flavour)

    async def _select_from(
        self,
        parent_path: 'FilePath', 
        is_dir: CoroutineMethod, 
        exists: CoroutineMethod, 
        scandir: CoroutineMethod
    ) -> AsyncIterable['FilePath']:
        try:
            path = parent_path._make_child_relpath(self.name)

            if await (is_dir if self.dironly else exists)(path):
                async for p in self.successor._select_from(path, is_dir, exists, scandir):
                    yield p
        except PermissionError:
            return


class _WildcardSelector(_AsyncSelector):
    def __init__(self, pat: str, child_parts: List[str], flavour: _Flavour):
        self.match = flavour.compile_pattern(pat)
        super().__init__(child_parts, flavour)

    async def _select_from(
        self,
        parent_path: 'FilePath', 
        is_dir: CoroutineMethod, 
        exists: CoroutineMethod, 
        scandir: CoroutineMethod
    ) -> AsyncIterable['FilePath']:
        try:
            async for entry in scandir(parent_path):
                if self.dironly:
                    try:
                        # "entry.is_dir()" can raise PermissionError
                        # in some cases (see bpo-38894), which is not
                        # among the errors ignored by _ignore_error()
                        if not await entry.is_dir(): continue
                    except OSError as e:
                        if not _ignore_error(e): raise
                        continue

                name = entry.name

                if self.match(name):
                    path = parent_path._make_child_relpath(name)

                async for p in self.successor._select_from(path, is_dir, exists, scandir):
                    yield p
        except PermissionError:
            return


class _RecursiveWildcardSelector(_AsyncSelector):
    def __init__(self, pat: str, child_parts: List[str], flavour):
        super().__init__(child_parts, flavour)

    async def _iterate_directories(
        self,
        parent_path: 'FilePath', 
        is_dir: CoroutineMethod, 
        scandir: CoroutineMethod
    ) -> AsyncIterable['FilePath']:
        yield parent_path

        try:
            async for entry in scandir(parent_path):
                entry_is_dir: bool = False

                try:
                    entry_is_dir = await entry.is_dir()

                except OSError as e:
                    if not _ignore_error(e):
                        raise

                if entry_is_dir and not await entry.is_symlink():
                    path = parent_path._make_child_relpath(entry.name)

                    async for p in self._iterate_directories(path, is_dir, scandir):
                        yield p
        except PermissionError:
            return

    async def _select_from(self, parent_path, is_dir, exists, scandir) -> AsyncIterable['FilePath']:
        try:
            yielded = set()

            try:
                successor_select = self.successor._select_from

                async for starting_point in self._iterate_directories(parent_path, is_dir, scandir):
                    async for p in successor_select(starting_point, is_dir, exists, scandir):
                        if p not in yielded:
                            yield p
                            yielded.add(p)

            finally:
                yielded.clear()
        except PermissionError:
            return


def _make_selector(pattern_parts: List[str], flavour: _Flavour) -> _AsyncSelector:
    pat: str = pattern_parts[0]
    child_parts: List[str] = pattern_parts[1:]

    if pat == '**':
        cls = _RecursiveWildcardSelector

    elif '**' in pat:
        raise ValueError("Invalid pattern: '**' can only be an entire path component")

    elif _is_wildcard_pattern(pat):
        cls = _WildcardSelector

    else:
        cls = _PreciseSelector

    return cls(pat, child_parts, flavour)


if hasattr(functools, "lru_cache"):
    _make_selector: Callable = functools.lru_cache()(_make_selector)


from __future__ import annotations

import io

from inspect import iscoroutinefunction
from contextlib import asynccontextmanager
from typing import AsyncContextManager
from anyio import AsyncFile, open_file
from aiofile import AIOFile, LineReader
from typing import AsyncIterable, Union, TYPE_CHECKING, Optional, cast, Tuple

from fileio.core._pathlib import Path
from fileio.types.common import Final, FileMode


if TYPE_CHECKING:  # keep mypy quiet
    from fileio.core.base import FilePath


BEGINNING: Final[int] = 0
CHUNK_SIZE: Final[int] = 4 * 1_024

SEP: Final[str] = '\n'
ENCODING: Final[str] = 'utf-8'
ERRORS: Final[str] = 'replace'


Paths = Union['FilePath', Path, str]
FileData = Union[bytes, str]


class IterableAIOFile(AIOFile):
    def __init__(
        self,
        *args,
        errors: Optional[str] = ERRORS,
        newline: Optional[str] = SEP,
        **kwargs
    ):
        super().__init__(*args, **kwargs)
        self._errors: Optional[str] = errors
        self._newline: Optional[str] = newline

        self._offset: int = 0

    def __aiter__(self) -> AsyncIterable[str]:
        encoding, errors, line_sep = self._get_options()

        return read_lines(
            self.name,
            line_sep,
            encoding=encoding,
            errors=errors,
        )

    def _set_offset(self, offset: int, data: FileData):
        self._offset = offset + len(data)

    def _get_options(
        self,
        encoding: Optional[str] = None,
        errors: Optional[str] = None
    ) -> Tuple[str, str, str]:
        encoding = encoding or self.encoding or ENCODING
        errors = errors or self._errors or ERRORS
        line_sep: str = self._newline or SEP

        return encoding, errors, line_sep

    async def read_text(
        self,
        encoding: Optional[str] = None,
        errors: Optional[str] = None
    ) -> str:
        encoding, errors, line_sep = self._get_options(encoding, errors)

        return await read_full_file(
            self.name,
            line_sep,
            encoding=encoding,
            errors=errors
        )

    async def read(
        self,
        size: int = -1,
        offset: Optional[int] = None
    ) -> FileData:
        if offset is None:
            offset = self._offset

        data: FileData = await super().read(size, offset)
        self._set_offset(offset, data)

        return data

    async def write(
        self,
        data: FileData,
        offset: Optional[int] = None
    ):
        if offset is None:
            offset = self._offset

        await super().write(data, offset)
        self._set_offset(offset, data)


async def read_lines(
    path: Paths,
    line_sep: str = SEP,
    chunk_size: int = CHUNK_SIZE,
    offset: int = BEGINNING,
    encoding: str = ENCODING,
    errors: str = ERRORS,
    **kwargs
    ) -> AsyncIterable[str]: 
    if hasattr(path, 'resolve'):
        if iscoroutinefunction(path.resolve):
            path = str(await path.resolve())

        else:
            path = str(path.resolve())

    path = cast(str, path)

    async with AIOFile(path, 'rb') as handle:
        reader = LineReader(
        handle,
        line_sep=line_sep,
        chunk_size=chunk_size,
        offset=offset
        )

        while True:
            line: bytes = await reader.readline()

            if not line: break

            yield line.decode(encoding, errors=errors)


async def read_full_file(
    path: Paths,
    line_sep: str = SEP,
    chunk_size: int = CHUNK_SIZE,
    offset: int = BEGINNING,
    encoding: str = ENCODING,
    errors: str = ERRORS,
    **kwargs
    ) -> str:
    lines_gen = read_lines(
        path,
        line_sep=line_sep,
        chunk_size=chunk_size,
        offset=offset,
        encoding=encoding,
        errors=errors
    )

    with io.StringIO() as string:
        async for line in lines_gen:
            string.write(line)

        return string.getvalue()


Handle = AsyncFile


@asynccontextmanager
async def get_handle(
    name: str,
    mode: FileMode = 'r',
    buffering: int = -1,
    encoding: str | None = ENCODING,
    errors: str | None = ERRORS,
    newline: str | None = SEP,
    ) -> AsyncContextManager[Handle]:
    file: AsyncFile

    if 'b' in mode:
        file = await open_file(name, mode)

    else:
        file = await open_file(
            name,
            mode,
            encoding=encoding,
            errors=errors,
            newline=newline,
        )

    yield file
    await file.aclose()



from pathlib import Path
from typing import List, Dict, Union, Any

try:
    import ruamel.yaml
    _yaml_available = True
except ImportError:
    _yaml_available = False


if _yaml_available:
    yaml = ruamel.yaml.YAML(typ=['rt', 'string'])
    yaml.preserve_quotes = True
    yaml.indent(
        mapping = 2, 
        sequence = 2, 
        offset = 0
    )


    class Yaml:

        @classmethod
        def dump(cls, obj: Dict[Any, Any], stream: Any = None, indent = 2, *args, **kwargs) -> Any:
            return yaml.dump(obj, stream = stream, indent = indent, *args, **kwargs)

        @classmethod
        def dumps(cls, obj: Dict[Any, Any], stream: Any = None, *args, **kwargs) -> Any:
            return yaml.dump_to_string(obj, *args, **kwargs)

        @classmethod
        def load(cls, data: Union[str, bytes, Any], *args, **kwargs) -> Union[Dict[Any, Any], List[str]]:
            return yaml.load(data, *args, **kwargs)
        
        @classmethod
        def loads_multi(cls, data: Union[str, bytes, Any], *args, _seperator: str = '\n---', **kwargs) -> Union[Dict[Any, Any], List[Dict[Any, Any]]]:
            """
            Loads a multi-yaml string into a list of Dict[Yaml].
            """
            _data = data.split(_seperator)
            try: return [yaml.load(d, *args, **kwargs) for d in _data]
            except Exception as e: 
                return yaml.load(data, *args, **kwargs)

        @classmethod
        def loads(cls, data: Union[str, bytes, Path, Any], *args, multi: bool = False, **kwargs) -> Union[Dict[Any, Any], List[str], List[Dict[Any, Any]]]:
            if not isinstance(data, str) and hasattr(data, 'as_posix'):
                # If it's a file-like object, read it as a string
                data = data.read_text()
            if '\n---' in data: multi = True 
            if multi: return cls.loads_multi(data, *args, **kwargs)
            return yaml.load(data, *args, **kwargs)



else:

    class Yaml:
        @classmethod
        def dump(cls, obj: Dict[Any, Any], stream: Any = None, indent = 2, *args, **kwargs) -> Any:
            raise ImportError("ruamel.yaml is not installed.")

        @classmethod
        def dumps(cls, obj: Dict[Any, Any], stream: Any = None, *args, **kwargs) -> Any:
            raise ImportError("ruamel.yaml is not installed.")

        @classmethod
        def load(cls, data: Union[str, bytes, Any], *args, **kwargs) -> Union[Dict[Any, Any], List[str]]:
            raise ImportError("ruamel.yaml is not installed.")
        
        @classmethod
        def loads_multi(cls, data: Union[str, bytes, Any], *args, _seperator: str = '\n---', **kwargs) -> Union[Dict[Any, Any], List[Dict[Any, Any]]]:
            """
            Loads a multi-yaml string into a list of Dict[Yaml].
            """
            raise ImportError("ruamel.yaml is not installed.")
            

        @classmethod
        def loads(cls, data: Union[str, bytes, Path, Any], *args, multi: bool = False, **kwargs) -> Union[Dict[Any, Any], List[str], List[Dict[Any, Any]]]:
            raise ImportError("ruamel.yaml is not installed.")
            



import base64
import gzip

from typing import Union
from fileio.io._base import BasePack


class Base64GZip(BasePack):
    encoding: str = 'utf-8'

    @classmethod
    def encode(cls, text: str, encoding: str = 'utf-8', *args, **kwargs) -> str:
        return base64.b64encode(
            gzip.compress(
                text.encode(encoding = encoding), *args, **kwargs)
            ).decode(encoding = encoding)

    @classmethod
    def decode(cls, data: Union[str, bytes], encoding: str = 'utf-8', *args, **kwargs) -> str:
        if isinstance(data, str): data = data.encode(encoding = encoding)
        return gzip.decompress(
            base64.b64decode(data, *args, **kwargs)
        ).decode(encoding = encoding)

    @classmethod
    def dumps(cls, data: str, encoding: str = 'utf-8', *args, **kwargs) -> str:
        return cls.encode(data, encoding = encoding, *args, **kwargs)
    
    @classmethod
    def loads(cls, data: Union[str, bytes], encoding: str = 'utf-8', *args, **kwargs) -> str:
        return cls.decode(data, encoding = encoding, *args, **kwargs)




from __future__ import absolute_import

from fileio.utils.logs import default_logger as logger
from fileio.utils.lazylib import LazyLib

from fileio.utils.helpers import *
from fileio.utils.configs import settings



"""
Handles Importing the correct pathlib for this library in
python3.10 cases
"""

import sys

# if 3.10
if sys.version_info.minor >= 10:
    from fileio.core import _pathlib as pathlib
    from fileio.core._pathlib import Path
else:
    import pathlib
    from pathlib import Path


__all__ = ('pathlib', 'Path')


import csv
from typing import Dict, Any, Union, List, Sequence, Generator, Iterator, Iterable, TYPE_CHECKING
from fileio.io._base import BasePack

if TYPE_CHECKING:
    from fileio.core.types import FileLike

class Tsv(BasePack):

    @classmethod
    def dumps(
        cls, 
        data: List[Dict[Any, Any]], 
        path: 'FileLike',
        mode: str = 'w',
        newline: str = '\n',
        buffering: int = -1,
        encoding: str = 'utf-8',
        fieldnames: Sequence = None, 
        delimiter: str = '\t',
        quotechar: str = '"',
        doublequote: bool = True,
        skipinitialspace: bool = False,
        lineterminator: bool = '\n',
        write_header: bool = True,
        **kwargs
    ) -> 'FileLike':
        from fileio import File
        file = File(path)
        with file.open(mode = mode, newline = newline, buffering = buffering, encoding = encoding) as f:
            fieldnames = fieldnames if fieldnames is not None else data[0].keys()
            writer = csv.DictWriter(
                f, 
                fieldnames = fieldnames,
                delimiter = delimiter,
                quotechar = quotechar,
                doublequote = doublequote,
                skipinitialspace = skipinitialspace,
                lineterminator = lineterminator,
                **kwargs
            )
            if write_header: writer.writeheader()
            for item in data:
                writer.writerow(item)
        return file
        

    @classmethod
    def load(
        cls, 
        path: 'FileLike', 
        mode: str = 'r', 
        newline: str = '\n',
        buffering: int = -1,
        encoding: str = 'utf-8',
        fieldnames: Sequence = None, 
        delimiter: str = '\t',
        quotechar: str = '"',
        doublequote: bool = True,
        skipinitialspace: bool = False,
        lineterminator: bool = '\n',
        **kwargs
    ) -> csv.DictReader:
        from fileio import File
        return csv.DictReader(
            File(path).open(mode = mode, newline = newline, buffering = buffering, encoding = encoding), 
            fieldnames = fieldnames, 
            delimiter = delimiter,
            quotechar = quotechar,
            doublequote = doublequote,
            skipinitialspace = skipinitialspace,
            lineterminator = lineterminator,
            **kwargs
        )
    
    @classmethod
    def loads(
        cls, 
        path: 'FileLike', 
        mode: str = 'r', 
        newline: str = '\n',
        buffering: int = -1,
        encoding: str = 'utf-8',
        fieldnames: Sequence = None, 
        delimiter: str = '\t',
        quotechar: str = '"',
        doublequote: bool = True,
        skipinitialspace: bool = False,
        lineterminator: bool = '\n',
        **kwargs
    ) -> Generator[Dict, None, None]:
        from fileio import File
        file = File(path)
        with file.open(mode = mode, newline = newline, buffering = buffering, encoding = encoding) as f:
            yield from csv.DictReader(
                f, 
                fieldnames = fieldnames, 
                delimiter = delimiter,
                quotechar = quotechar,
                doublequote = doublequote,
                skipinitialspace = skipinitialspace,
                lineterminator = lineterminator,
                **kwargs
            )






import json
import datetime
import dataclasses
import contextlib
from typing import Dict, Any, Union, List, Type

from fileio.io._base import BasePack

try:
    import numpy as np
except ImportError:
    np = None



def object_serializer(obj: Any) -> Any:
    if isinstance(obj, dict):
        return {k: object_serializer(v) for k, v in obj.items()}

    if isinstance(obj, bytes):
        return obj.decode('utf-8')

    if isinstance(obj, (str, list, dict, int, float, bool, type(None))):
        return obj

    if dataclasses.is_dataclass(obj):
        return dataclasses.asdict(obj)

    if hasattr(obj, 'dict'): # test for pydantic models
        return obj.dict()
    
    if hasattr(obj, 'get_secret_value'):
        return obj.get_secret_value()
    
    if hasattr(obj, 'as_posix'):
        return obj.as_posix()
    
    if hasattr(obj, "numpy"):  # Checks for TF tensors without needing the import
        return obj.numpy().tolist()
    
    if hasattr(obj, 'tolist'): # Checks for torch tensors without importing
        return obj.tolist()
    
    if isinstance(obj, (datetime.date, datetime.datetime)):
        return obj.isoformat()
    
    if np is not None:
        if isinstance(obj, (np.int_, np.intc, np.intp, np.int8, np.int16, np.int32, np.int64, np.uint8, np.uint16, np.uint32, np.uint64)):
            return int(obj)
        
        if isinstance(obj, (np.float_, np.float16, np.float32, np.float64)):
            return float(obj)
        
    else:
        # Try to convert to a primitive type
        with contextlib.suppress(Exception):
            return int(obj)
        with contextlib.suppress(Exception):
            return float(obj)

    raise TypeError(f"Object of type {type(obj)} is not JSON serializable")


class ObjectEncoder(json.JSONEncoder):
    
    def default(self, obj: Any):   # pylint: disable=arguments-differ,method-hidden
        try:
            return object_serializer(obj)
        except Exception as e:
            return json.JSONEncoder.default(self, obj)

class Json:

    @staticmethod
    def dumps(obj: Dict[Any, Any], *args, default: Dict[Any, Any] = None, cls: Type[json.JSONEncoder] = ObjectEncoder, **kwargs) -> str:
        return json.dumps(obj, *args, default = default, cls = cls, **kwargs)

    @staticmethod
    def loads(data: Union[str, bytes], *args, **kwargs) -> Union[Dict[Any, Any], List[str]]:
        return json.loads(data, *args, **kwargs)


try:
    import orjson
    class OrJson(BasePack):
        
        @classmethod
        def dumps(cls, obj: Dict[Any, Any], *args, default: Dict[Any, Any] = None, **kwargs) -> str:
            return orjson.dumps(obj, default=default, *args, **kwargs).decode()
        
        @classmethod
        def loads(cls, data: Union[str, bytes], *args, **kwargs) -> Union[Dict[Any, Any], List[str]]:
            return orjson.loads(data, *args, **kwargs)


except ImportError:
    # Use Default JSON if not installed
    class OrJson(Json):
        pass


